@@ -301,9 +301,9 @@ errout:
 }
 
 errcode_t
-profile_init(const char **files, profile_t *ret_profile)
+profile_init(const char * const *files, profile_t *ret_profile)
 {
-	const char **fs;
+	const char * const *fs;
 	profile_t profile;
 	prf_file_t  new_file, *last;
 	errcode_t retval = 0;
@@ -1112,7 +1112,7 @@ errcode_t profile_create_node(const char *name, const char *value,
 }
 
 /*
- * This function verifies that all of the representation invarients of
+ * This function verifies that all of the representation invariants of
  * the profile are true.  If not, we have a programming bug somewhere,
  * probably in this file.
  */
@@ -1362,7 +1362,7 @@ errcode_t profile_node_iterator(void **iter_p, struct profile_node **ret_node,
 	}
 get_new_file:
 	if (iter->node == 0) {
-		if (iter->file == 0 ||
+		if (iter->file == NULL ||
 		    (iter->flags & PROFILE_ITER_FINAL_SEEN)) {
 			profile_iterator_free(iter_p);
 			if (ret_node)
@@ -1376,7 +1376,8 @@ get_new_file:
 		if ((retval = profile_update_file(iter->file))) {
 		    if (retval == ENOENT || retval == EACCES) {
 			/* XXX memory leak? */
-			iter->file = iter->file->next;
+			if (iter->file)
+			    iter->file = iter->file->next;
 			skip_num = 0;
 			retval = 0;
 			goto get_new_file;
@@ -1405,7 +1406,8 @@ get_new_file:
 				iter->flags |= PROFILE_ITER_FINAL_SEEN;
 		}
 		if (!section) {
-			iter->file = iter->file->next;
+			if (iter->file)
+				iter->file = iter->file->next;
 			skip_num = 0;
 			goto get_new_file;
 		}
@@ -1435,13 +1437,15 @@ get_new_file:
 	}
 	iter->num++;
 	if (!p) {
-		iter->file = iter->file->next;
+		if (iter->file)
+			iter->file = iter->file->next;
 		iter->node = 0;
 		skip_num = 0;
 		goto get_new_file;
 	}
 	if ((iter->node = p->next) == NULL)
-		iter->file = iter->file->next;
+		if (iter->file)
+			iter->file = iter->file->next;
 	if (ret_node)
 		*ret_node = p;
 	if (ret_name)
